@@ -1,6 +1,5 @@
 # Licensed under a 3-clause BSD style license - see LICENSE.rst
 import pytest
-import numpy as np
 from numpy.testing import assert_allclose
 from .test_fit import MyDataset
 from ..datasets import Datasets
@@ -8,7 +7,7 @@ from ..datasets import Datasets
 
 @pytest.fixture(scope="session")
 def datasets():
-    return Datasets([MyDataset(), MyDataset()], mask=np.array([True]))
+    return Datasets([MyDataset(), MyDataset()])
 
 
 class TestDatasets:
